// contactmodelEEPA_LIN.cpp
#include "eepa_lin.h"

#include "module/interface/icontactmechanical.h"
#include "module/interface/icontact.h"
#include "module/interface/ipiecemechanical.h"
#include "module/interface/ipiece.h"
#include "module/interface/ifishcalllist.h"

#include "version.txt"

#include "utility/src/tptr.h"
#include "shared/src/mathutil.h"

#include "kernel/interface/iprogram.h"
#include "module/interface/icontactthermal.h"
#include "contactmodel/src/contactmodelthermal.h"

#ifdef EEPA_LIN_LIB
int __stdcall DllMain(void *, unsigned, void *) {
	return 1;
}

extern "C" EXPORT_TAG const char *getName() {
#if DIM==3
	return "contactmodelmechanical3dEEPA_LIN";
#else
	return "contactmodelmechanical2dEEPA_LIN";
#endif
}

extern "C" EXPORT_TAG unsigned getMajorVersion() {
	return MAJOR_VERSION;
}

extern "C" EXPORT_TAG unsigned getMinorVersion() {
	return MINOR_VERSION;
}

extern "C" EXPORT_TAG void *createInstance() {
	cmodelsxd::ContactModelEEPA_LIN *m = new cmodelsxd::ContactModelEEPA_LIN();
	return (void *)m;
}
#endif 

namespace cmodelsxd {
	static const quint32 shearMask = 0x00000002; // Base 1!
	static const quint32 poissMask = 0x00000004;
	static const quint32 fricMask = 0x00000008;
	static const quint32 resFricMask = 0x00004000;

	using namespace itasca;

	int ContactModelEEPA_LIN::index_ = -1;
	UInt ContactModelEEPA_LIN::getMinorVersion() const { return MINOR_VERSION; }

	ContactModelEEPA_LIN::ContactModelEEPA_LIN() : inheritanceField_(shearMask | poissMask | fricMask | resFricMask)
		, effectiveTranslationalStiffness_(DVect2(0.0))
		, effectiveRotationalStiffness_(DAVect(0.0))
		, shear_(0.0)
		, poiss_(0.0)
		, fric_(0.0)
		, eepa_F_(DVect(0.0))
		, eepa_S_(false)
		, rgap_(0.0)
		, dpProps_(0)
		, res_fric_(0.0)
		, res_M_(DAVect(0.0))
		, res_S_(false)
		, kr_(0.0)
		, fr_(0.0)
		, Moverlap_(0.0)			
		, Poverlap_(0.0)
		, plas_ratio_(0.5)
		, pull_off_(0.0)
		, surf_adh_(0.0)
		, k1_(0.0)			
		, k2_(0.0)			
		, kadh_(0.0)
		, ks_(0.0)					
		, r_hertz_(0.0)
		, ks_fac_(1.0)
		, f_min_(0.0)
		, branch_(1)
		, energies_(0) {
	}

	ContactModelEEPA_LIN::~ContactModelEEPA_LIN() {
		// Make sure to clean up after yourself!
		if (dpProps_)
			delete dpProps_;
		if (energies_)
			delete energies_;
	}

	void ContactModelEEPA_LIN::archive(ArchiveStream &stream) {
		// The stream allows one to archive the values of the contact model
		// so that it can be saved and restored. The minor version can be
		// used here to allow for incremental changes to the contact model too. 
		stream & shear_;
		stream & poiss_;
		stream & fric_;
		stream & eepa_F_;
		stream & eepa_S_;
		stream & rgap_;
		stream & res_fric_;
		stream & res_M_;
		stream & res_S_;
		stream & kr_;
		stream & fr_;
		stream & Moverlap_;									
		stream & Poverlap_;
		stream & plas_ratio_;
		stream & pull_off_;
		stream & surf_adh_;
		stream & k1_;			
		stream & k2_;			
		stream & kadh_;
		stream & ks_;										
		stream & r_hertz_;
		stream & ks_fac_;
		stream & f_min_;
		stream & branch_;

		if (stream.getArchiveState() == ArchiveStream::Save) {
			bool b = false;
			if (dpProps_) {
				b = true;
				stream & b;
				stream & dpProps_->dp_nratio_;
				stream & dpProps_->dp_sratio_;
				stream & dpProps_->dp_mode_;
				stream & dpProps_->dp_F_;
			}
			else
				stream & b;

			b = false;
			if (energies_) {
				b = true;
				stream & b;
				stream & energies_->estrain_;
				stream & energies_->errstrain_;
				stream & energies_->eslip_;
				stream & energies_->errslip_;
				stream & energies_->edashpot_;
			}
			else
				stream & b;
		}
		else {
			bool b(false);
			stream & b;
			if (b) {
				if (!dpProps_)
					dpProps_ = NEWC(dpProps());
				stream & dpProps_->dp_nratio_;
				stream & dpProps_->dp_sratio_;
				stream & dpProps_->dp_mode_;
				stream & dpProps_->dp_F_;
			}
			stream & b;
			if (b) {
				if (!energies_)
					energies_ = NEWC(Energies());
				stream & energies_->estrain_;
				stream & energies_->errstrain_;
				stream & energies_->eslip_;
				stream & energies_->errslip_;
				stream & energies_->edashpot_;
			}
		}

		stream & inheritanceField_;
		stream & effectiveTranslationalStiffness_;
		stream & effectiveRotationalStiffness_;

	}

	void ContactModelEEPA_LIN::copy(const ContactModel *cm) {
		// Copy all of the contact model properties. Used in the CMAT 
		// when a new contact is created. 
		ContactModelMechanical::copy(cm);
		const ContactModelEEPA_LIN *in = dynamic_cast<const ContactModelEEPA_LIN*>(cm);
		if (!in) throw std::runtime_error("Internal error: contact model dynamic cast failed.");
		shear(in->shear());
		poiss(in->poiss());
		fric(in->fric());
		eepa_F(in->eepa_F());
		eepa_S(in->eepa_S());
		rgap(in->rgap());
		res_fric(in->res_fric());
		res_M(in->res_M());
		res_S(in->res_S());
		kr(in->kr());
		fr(in->fr());
		Moverlap(in->Moverlap());					
		Poverlap(in->Poverlap());
		plas_ratio(in->plas_ratio());
		pull_off(in->pull_off());
		surf_adh(in->surf_adh());
		k1(in->k1());			
		k2(in->k2());			
		kadh(in->kadh());
		ks(in->ks());	
		r_hertz(in->r_hertz());
		ks_fac(in->ks_fac());
		f_min(in->f_min());
		branch(in->branch());

		if (in->hasDamping()) {
			if (!dpProps_)
				dpProps_ = NEWC(dpProps());
			dp_nratio(in->dp_nratio());
			dp_sratio(in->dp_sratio());
			dp_mode(in->dp_mode());
			dp_F(in->dp_F());
		}
		if (in->hasEnergies()) {
			if (!energies_)
				energies_ = NEWC(Energies());
			estrain(in->estrain());
			errstrain(in->errstrain());
			eslip(in->eslip());
			errslip(in->errslip());
			edashpot(in->edashpot());
		}
		inheritanceField(in->inheritanceField());
		effectiveTranslationalStiffness(in->effectiveTranslationalStiffness());
		effectiveRotationalStiffness(in->effectiveRotationalStiffness());
	}

	QVariant ContactModelEEPA_LIN::getProperty(uint i, const IContact *con) const {
		// Return the property. The IContact pointer is provided so that 
		// more complicated properties, depending on contact characteristics,
		// can be calcualted. 
		QVariant var;
		switch (i) {
		case kwShear:     return shear_;
		case kwPoiss:     return poiss_;
		case kwFric:      return fric_;
		case kwEepaF:     var.setValue(eepa_F_); return var;
		case kwEepaS:     return eepa_S_;
		case kwRGap:      return rgap_;
		case kwDpNRatio:  return dpProps_ ? dpProps_->dp_nratio_ : 0;
		case kwDpSRatio:  return dpProps_ ? dpProps_->dp_sratio_ : 0;
		case kwDpMode:    return dpProps_ ? dpProps_->dp_mode_ : 0;
		case kwDpF: {
			dpProps_ ? var.setValue(dpProps_->dp_F_) : var.setValue(DVect(0.0));
			return var;
		}
		case kwResFric:    return res_fric_;
		case kwResMoment:  var.setValue(res_M_); return var;
		case kwResS:       return res_S_;
		case kwMoverlap:   return Moverlap_;														
		case kwPlasRat:	   return plas_ratio_;
		case kwPullOff:	   return pull_off_;
		case kwSurfAdh:	   return surf_adh_;
		case kwKsFac:	   return ks_fac_;															
		case kwFmin:	   return f_min_;
		}
		assert(0);
		return QVariant();
	}

	bool ContactModelEEPA_LIN::getPropertyGlobal(uint i) const {
		// Returns whether or not a property is held in the global axis system (TRUE)
		// or the local system (FALSE). Used by the plotting logic.
		switch (i) {
		case kwEepaF:
		case kwDpF:
		case kwResMoment:
			return false;
		}
		return true;
	}

	bool ContactModelEEPA_LIN::setProperty(uint i, const QVariant &v, IContact *) {
		// Set a contact model property. Return value indicates that the timestep
		// should be recalculated. 
		dpProps dp;
		switch (i) {
		case kwShear: {
			if (!v.canConvert<double>())
				throw Exception("eepa_shear must be a double.");
			double val(v.toDouble());
			if (val <= 0.0)
				throw Exception("zero or negative shear not allowed.");
			shear_ = val;
			return true;
		}
		case kwPoiss: {
			if (!v.canConvert<double>())
				throw Exception("eepa_poiss must be a double.");
			double val(v.toDouble());
			if (val < 0.0 || val > 0.5)
				throw Exception("eepa_poiss must be in the range [0, 0.5].");
			poiss_ = val;
			return true;
		}
		case kwFric: {
			if (!v.canConvert<double>())
				throw Exception("fric must be a double.");
			double val(v.toDouble());
			if (val < 0.0)
				throw Exception("negative fric not allowed.");
			fric_ = val;
			return false;
		}
		case kwEepaF: {
			if (!v.canConvert<DVect>())
				throw Exception("eepa_force must be a vector.");
			DVect val(v.value<DVect>());
			eepa_F_ = val;
			return false;
		}
		case kwRGap: {
			if (!v.canConvert<double>())
				throw Exception("reference gap must be a double.");
			double val(v.toDouble());
			rgap_ = val;
			return false;
		}
		case kwDpNRatio: {
			if (!v.canConvert<double>())
				throw Exception("dp_nratio must be a double.");
			double val(v.toDouble());
			if (val < 0.0)
				throw Exception("negative dp_nratio not allowed.");
			if (val == 0.0 && !dpProps_)
				return false;
			if (!dpProps_)
				dpProps_ = NEWC(dpProps());
			dpProps_->dp_nratio_ = val;
			return true;
		}
		case kwDpSRatio: {
			if (!v.canConvert<double>())
				throw Exception("dp_sratio must be a double.");
			double val(v.toDouble());
			if (val < 0.0)
				throw Exception("negative dp_sratio not allowed.");
			if (val == 0.0 && !dpProps_)
				return false;
			if (!dpProps_)
				dpProps_ = NEWC(dpProps());
			dpProps_->dp_sratio_ = val;
			return true;
		}
		case kwDpMode: {
			if (!v.canConvert<int>())
				throw Exception("the viscous mode dp_mode must be 0 or 1");
			int val(v.toInt());
			if (val == 0 && !dpProps_)
				return false;
			if (val < 0 || val > 1)
				throw Exception("the viscous mode dp_mode must be 0 or 1.");
			if (!dpProps_)
				dpProps_ = NEWC(dpProps());
			dpProps_->dp_mode_ = val;
			return false;
		}
		case kwDpF: {
			if (!v.canConvert<DVect>())
				throw Exception("dp_force must be a vector.");
			DVect val(v.value<DVect>());
			if (val.fsum() == 0.0 && !dpProps_)
				return false;
			if (!dpProps_)
				dpProps_ = NEWC(dpProps());
			dpProps_->dp_F_ = val;
			return false;
		}
		case kwResFric: {
			if (!v.canConvert<double>())
				throw Exception("res_fric must be a double.");
			double val(v.toDouble());
			if (val < 0.0)
				throw Exception("negative res_fric not allowed.");
			res_fric_ = val;
			return true;
		}
		case kwResMoment: {
			DAVect val(0.0);
#ifdef TWOD               
			if (!v.canConvert<DAVect>() && !v.canConvert<double>())
				throw Exception("res_moment must be an angular vector.");
			if (v.canConvert<DAVect>())
				val = DAVect(v.value<DAVect>());
			else
				val = DAVect(v.toDouble());
#else
			if (!v.canConvert<DAVect>() && !v.canConvert<DVect>())
				throw Exception("res_moment must be an angular vector.");
			if (v.canConvert<DAVect>())
				val = DAVect(v.value<DAVect>());
			else
				val = DAVect(v.value<DVect>());
#endif
			res_M_ = val;
			return false;
		}				
		case kwPlasRat: {																				
			if (!v.canConvert<double>())
				throw Exception("plas_rat must be a double.");
			 double val(v.toDouble());
			 if (val <= 0.0 || val >= 1.0)
			     throw Exception("plas_rat must be larger than zero and smaller than 1.");
			 plas_ratio_ = val;
			 return true;
		}
		case kwPullOff: {
			if (!v.canConvert<double>())
				throw Exception("constant pull-off force must be a double.");
			double val(v.toDouble());
			if (val > 0.0)
				throw Exception("pull_off must be a negative value or zero.");
			pull_off_ = val;
			return false;
		}
		case kwSurfAdh: {
			if (!v.canConvert<double>())
				throw Exception("surface adhesion energy must be a double.");
			double val(v.toDouble());
			if (val < 0.0)
				throw Exception("negative surf_adh not allowed.");
			surf_adh_ = val;
			return false;
		}																								
		case kwKsFac: {
			if (!v.canConvert<double>())
				throw Exception("shear stiffness factor must be a double.");
			double val(v.toDouble());
			if (val <= 0.0)
				throw Exception("zero or negative ks_fac not allowed.");
			ks_fac_ = val;
			return true;
		}																								

		}//switch
		return false;
	}

	bool ContactModelEEPA_LIN::getPropertyReadOnly(uint i) const {
		// Returns TRUE if a property is read only or FALSE otherwise. 
		switch (i) {
		case kwDpF:
		case kwEepaS:
		case kwResS:
		case kwMoverlap:	
		case kwFmin:
			return true;
		default:
			break;
		}
		return false;
	}

	bool ContactModelEEPA_LIN::supportsInheritance(uint i) const {
		// Returns TRUE if a property supports inheritance or FALSE otherwise. 
		switch (i) {
		case kwShear:
		case kwPoiss:
		case kwFric:
		case kwResFric:
			return true;
		default:
			break;
		}
		return false;
	}

	double ContactModelEEPA_LIN::getEnergy(uint i) const {
		// Return an energy value. 
		double ret(0.0);
		if (!energies_)
			return ret;
		switch (i) {
		case kwEStrain:    return energies_->estrain_;
		case kwERRStrain:  return energies_->errstrain_;
		case kwESlip:      return energies_->eslip_;
		case kwERRSlip:    return energies_->errslip_;
		case kwEDashpot:   return energies_->edashpot_;
		}
		assert(0);
		return ret;
	}

	bool ContactModelEEPA_LIN::getEnergyAccumulate(uint i) const {
		// Returns TRUE if the corresponding energy is accumulated or FALSE otherwise.
		switch (i) {
		case kwEStrain:   return true;
		case kwERRStrain: return false;
		case kwESlip:     return true;
		case kwERRSlip:   return true;
		case kwEDashpot:  return true;
		}
		assert(0);
		return false;
	}

	void ContactModelEEPA_LIN::setEnergy(uint i, const double &d) {
		// Set an energy value. 
		if (!energies_) return;
		switch (i) {
		case kwEStrain:    energies_->estrain_ = d;   return;
		case kwERRStrain:  energies_->errstrain_ = d; return;
		case kwESlip:      energies_->eslip_ = d;   return;
		case kwERRSlip:    energies_->errslip_ = d; return;
		case kwEDashpot:   energies_->edashpot_ = d;   return;
		}
		assert(0);
		return;
	}

	bool ContactModelEEPA_LIN::validate(ContactModelMechanicalState *state, const double &) {
		// Validate the / Prepare for entry into ForceDispLaw. The validate function is called when:
		// (1) the contact is created, (2) a property of the contact that returns a true via
		// the setProperty method has been modified and (3) when a set of cycles is executed
		// via the {cycle N} command.
		// Return value indicates contact activity (TRUE: active, FALSE: inactive).
		assert(state);
		const IContactMechanical *c = state->getMechanicalContact();
		assert(c);
		//
		if (state->trackEnergy_)
			activateEnergy();
		//		
		if ((inheritanceField_ & shearMask) || (inheritanceField_ & poissMask))
			updateEndStiffCoef(c);
		if (inheritanceField_ & fricMask)
			updateEndFric(c);
		if (inheritanceField_ & resFricMask)
			updateEndResFric(c);
		//
		if (shear_ <= 0.0) {
			throw Exception("'eepa_shear' must be specified using a value larger than zero in the 'eepa_lin' model.");
		}
		//
		updateStiffCoef(c);								// calculate the stiffness values based on the material properties - specified directly or via inheritance
		updateEffectiveStiffness(state);				// effective stiffness for translation and rotation used in the time step estimation
		//
		return checkActivity(state->gap_);
	}

	void ContactModelEEPA_LIN::updateStiffCoef(const IContactMechanical *con) {
		//
		double c12 = con->getEnd1Curvature().y();
		double c22 = con->getEnd2Curvature().y();
		r_hertz_ = c12 + c22;
		if (r_hertz_ == 0.0)
			throw Exception("eepa_lin contact model undefined for 2 non-curved surfaces");
		r_hertz_ = 1.0 / r_hertz_;						// R1*R2/(R1 + R2)
		k1_ = 2.0*(shear_ / (1 - poiss_)) * r_hertz_;	// k1 loading branch
		k2_ = k1_ / (1.0 - plas_ratio_);				// k2 loading branch
		ks_ = ks_fac_ * k1_;							// shear stiffness
		// rolling resistance 
		kr_ = 0.0;
		fr_ = 0.0;
		if (res_fric_ > 0.0) {
			fr_ = res_fric_ * r_hertz_;						// store this value with contact - used in force-displacement calculation
			kr_ = ks_ * r_hertz_ * r_hertz_;				// based on 'ks_'
		}
	}

	bool ContactModelEEPA_LIN::endPropertyUpdated(const QString &name, const IContactMechanical *c) {
		// The endPropertyUpdated method is called whenever a surface property (with a name
		// that matches an inheritable contact model property name) of one of the contacting
		// pieces is modified. This allows the contact model to update its associated
		// properties. The return value denotes whether or not the update has affected
		// the time step computation (by having modified the translational or rotational
		// tangent stiffnesses). If true is returned, then the time step will be recomputed.  
		assert(c);
		QStringList availableProperties = getProperties().simplified().replace(" ", "").split(",", QString::SkipEmptyParts);
		QRegExp rx(name, Qt::CaseInsensitive);
		int idx = availableProperties.indexOf(rx) + 1;
		bool ret = false;

		if (idx <= 0)
			return ret;

		switch (idx) {
		case kwShear: { //Shear
			if (inheritanceField_ & shearMask)
				ret = true;							// return 'true' to ensure 'validate()' is called to affect the change in Shear Modulus through 'updateEndStiffCoef()' and effective (tangent) stiffnesses
			break;
		}
		case kwPoiss: { //Poisson's
			if (inheritanceField_ & poissMask)
				ret =true;							// return 'true' to ensure 'validate()' is called to affect the change in Poisson's ratio through 'updateEndStiffCoef()' and effective (tangent) stiffnesses
			break;
		}
		case kwFric: { //fric
			if (inheritanceField_ & fricMask)
				updateEndFric(c);					// friction does not influence any of the other parameters or time step size, so directly update the contact model friction
			break;
		}
		case kwResFric: { //rr_fric
			if (inheritanceField_ & resFricMask)
				ret = true;							// return 'true' to ensure 'validate()' is called to affect the change in rolling friction through 'updateEndResFric()' and 'fr_'
			break;
		}
		//
		}
		return ret;
	}

	static const QString gstr("eepa_shear");
	static const QString nustr("eepa_poiss");
	bool ContactModelEEPA_LIN::updateEndStiffCoef(const IContactMechanical *con) {
		assert(con);
		double g1 = shear_;
		double g2 = shear_;
		double nu1 = poiss_;
		double nu2 = poiss_;
		QVariant vg1 = con->getEnd1()->getProperty(gstr);
		QVariant vg2 = con->getEnd2()->getProperty(gstr);
		QVariant vnu1 = con->getEnd1()->getProperty(nustr);
		QVariant vnu2 = con->getEnd2()->getProperty(nustr);
		if (vg1.isValid() && vg2.isValid()) {
			g1 = vg1.toDouble();
			g2 = vg2.toDouble();
			if (g1 <= 0.0 || g2 <= 0.0)
				throw Exception("Negative or zero shear modulus not allowed in eepa_lin contact model");
		}
		if (vnu1.isValid() && vnu2.isValid()) {
			nu1 = vnu1.toDouble();
			nu2 = vnu2.toDouble();
			if (nu1 < 0.0 || nu1 > 0.5 || nu2 < 0.0 || nu2 > 0.5)
				throw Exception("Poisson ratio should be in range [0,0.5] in eepa_lin contact model");
		}
		if (g1*g2 == 0.0) return false;
		double es = 1.0 / ((1.0 - nu1) / (2.0*g1) + (1.0 - nu2) / (2.0*g2));
		double gs = 1.0 / ((2.0 - nu1) / g1 + (2.0 - nu2) / g2);
		poiss_ = (4.0*gs - es) / (2.0*gs - es);
		shear_ = 2.0*gs*(2 - poiss_);
		if (shear_ <= 0.0)
			throw Exception("Negative or zero shear modulus not allowed in eepa_lin contact model");
		if (poiss_ < 0.0 || poiss_ > 0.5)
			throw Exception("Poisson ratio should be in range [0,0.5] in eepa_lin contact model");
		return true;
	}

	static const QString fricstr("fric");
	bool ContactModelEEPA_LIN::updateEndFric(const IContactMechanical *con) {
		assert(con);
		QVariant v1 = con->getEnd1()->getProperty(fricstr);
		QVariant v2 = con->getEnd2()->getProperty(fricstr);
		if (!v1.isValid() || !v2.isValid())
			return false;
		double fric1 = std::max(0.0, v1.toDouble());
		double fric2 = std::max(0.0, v2.toDouble());
		double val = fric_;
		fric_ = std::min(fric1, fric2);
		if (fric_ <= 0.0)
			throw Exception("Negative friction not allowed in eepa_lin contact model");
		return ((fric_ != val));
	}

	static const QString rfricstr("rr_fric");
	bool ContactModelEEPA_LIN::updateEndResFric(const IContactMechanical *con) {
		assert(con);
		QVariant v1 = con->getEnd1()->getProperty(rfricstr);
		QVariant v2 = con->getEnd2()->getProperty(rfricstr);
		if (!v1.isValid() || !v2.isValid())
			return false;
		double rfric1 = std::max(0.0, v1.toDouble());
		double rfric2 = std::max(0.0, v2.toDouble());
		double val = res_fric_;
		res_fric_ = std::min(rfric1, rfric2);
		if (res_fric_ <= 0.0)
			throw Exception("Negative rolling friction not allowed in eepa_lin contact model");
		return ((res_fric_ != val));
	}
	
	void ContactModelEEPA_LIN::updateEffectiveStiffness(ContactModelMechanicalState *state) {
		//
		double overlap = rgap_ - state->gap_;
		//
		double kn_tangent = 0.0;
		double ks_tangent = ks_;
		//
		if (overlap <= 0.0) {					// inactive contact
			kn_tangent = k1_;
		}
		else {									// active contact
			switch (branch_) {
			case 1: {
				kn_tangent = k1_;
				break;
			}
			case 2: {
				kn_tangent = k2_;
				break;
			}
			case 3: {
				kn_tangent = kadh_;
				break;
			}
			default: {
				kn_tangent = k2_;				// k2_ > k1_ always
			}
			}//switch
		}
		//
		DVect2 retT(kn_tangent, ks_tangent);
		// correction if viscous damping active
		if (dpProps_) {
			DVect2 correct(1.0);
			if (dpProps_->dp_nratio_)
				correct.rx() = sqrt(1.0 + dpProps_->dp_nratio_*dpProps_->dp_nratio_) - dpProps_->dp_nratio_;
			if (dpProps_->dp_sratio_)
				correct.ry() = sqrt(1.0 + dpProps_->dp_sratio_*dpProps_->dp_sratio_) - dpProps_->dp_sratio_;
			retT /= (correct*correct);
		}
		//
		effectiveTranslationalStiffness_ = retT;									// set
		effectiveRotationalStiffness_ = DAVect(kr_);								// Effective rotational stiffness (bending only)
#if DIM==3 
		effectiveRotationalStiffness_.rx() = 0.0;
#endif
	}

	bool ContactModelEEPA_LIN::forceDisplacementLaw(ContactModelMechanicalState *state, const double &timestep) {
		assert(state);
		//
		// Current overlap
		double overlap = rgap_ - state->gap_;
		// Relative translational increment
		DVect trans = state->relativeTranslationalIncrement_;
		//
		// The contact was just activated from an inactive state
		if (state->activated()) {
			//
			eepa_F_.rdof(0) = pull_off_;												// the contact just got activated, set the initial normal force equal to the pull-off force
			//
			// Trigger the FISH callback if one is hooked up to the 
			// contact_activated event.
			if (cmEvents_[fActivated] >= 0) {
				// An FArray of QVariant is returned and these will be passed
				// to the FISH function as an array of FISH symbols as the second
				// argument to the FISH callback function. 
				FArray<QVariant, 2> arg;
				QVariant v;
				// Just put a pointer to the contact in the array and return it.
				IContact * c = const_cast<IContact*>(state->getContact());
				TPtr<IThing> t(c->getIThing());
				v.setValue(t);
				arg.push_back(v);
				IFishCallList *fi = const_cast<IFishCallList*>(state->getProgram()->findInterface<IFishCallList>());
				fi->setCMFishCallArguments(c, arg, cmEvents_[fActivated]);
			}
		}
		//
		// Angular dispacement increment.
		DAVect ang = state->relativeAngularIncrement_;
		//
		// EEPA NORMAL FORCE ==============================================================================================================	
		//
		double eepa_Fn = eepa_F_.dof(0);										// current normal force value
		DVect eepa_F_old = eepa_F_;												// store for energy computations only
		//
		// The updates below only occur when the maximum overlap is exceeded
		if (overlap > Moverlap_) {												// the current overlap is exceeding the maximum it has ever been for loading and unloading
			Moverlap_ = overlap;												// update the maximum contact overlap
			//
			double radius = sqrt(Poverlap_*2.0*r_hertz_);						// contact patch radius at plastic overlap   
			f_min_ = pull_off_ - 1.5*M_PI*surf_adh_ *radius;					// update the minimum force based on the surface adhesion energy and the current contact surface area
			//
			double f_min_limit = eepa_Fn - k2_ * overlap;						// the limit of Fmin that can be reached at zero overlap for the current force value
			if (f_min_ < f_min_limit) {
				f_min_ = 0.5*(pull_off_ + f_min_limit);							// if Fmin is less than the limit, set it to half the limit (Morissey, 2013)
			}
			//
			double overlap_min = overlap - (eepa_Fn - f_min_) / k2_;			// the minimum overlap raised to the power lu_exp - the overlap where the k2 unloading branch goes over into the adhesion branch
			kadh_ = (pull_off_ - f_min_) / overlap_min;							// stiffness of the adhesion branch	
		}
		//
		double F1 = pull_off_ + k1_ * overlap;									// k1 loading branch force for this overlap
		double F2 = pull_off_ + k2_ * (overlap - Poverlap_);					// k2 loading-unloading branch force for this overlap
		double Fadh = pull_off_ - kadh_ * overlap;								// adhesion branch force for this overlap
		//
		double kn_branch = 0.0;													// current branch stiffness in the normal direction - used for time step and damping calculation
		//		
		if (F2 >= F1) {															// the k1 loading branch should be followed
			eepa_Fn = F1;														// set the EEPA force to that of the k1 loading branch
			Poverlap_ = overlap - (F1 - pull_off_) / k2_;						// for the current force (F1), calculate the plastic overlap [not the usual equation based on maximum overlap, but the same result]  
			kn_branch = k1_;													// set the stiffness in the normal direction
			branch_ = 1;
		}
		else {																	// will only enter if F2 < F1
			if (F2 > Fadh) {													// on the k2 loading-unloading branch
				eepa_Fn = F2;													// set the EEPA force to that of the k2 loading-unloading branch
				kn_branch = k2_; 												// set the stiffness in the normal direction
				branch_ = 2;
			}
			else {																// only enter if F2 < F1 AND F2 < Fadh
				if (trans.x() >= 0.0) {											// contact is unloading along the adhesive branch (can not load along this branch)
					eepa_Fn = Fadh;												// set the EEPA force to that of the adhesive unloading branch
					kn_branch = kadh_;											// set the stiffness in the normal direction
					branch_ = 3;
				}
				else {															// loading on the adhesive branch which is not defined
					eepa_Fn -= k2_ * trans.x();									// set the EEPA force by incrementing the force from the previous time-step along the k2 loading branch. 
																				// The next time step, the updated Poverlap_exp is used and the code will automatically follow the k2 loading branch since F2 is calculated from this updated Poverlap_exp 
																				// The negative of the increment should be used to ensure a positive value is rasied to the power "lu_exp". The sign is then corrected by adding the force increment to the previous value
					Poverlap_ = overlap + (pull_off_ - eepa_Fn) / k2_;			// update the plastic overlap which means that the k2 loading-unloading branch has shifted in the unloading direction to start at this point
					kn_branch = k2_;											// set the stiffness in the normal direction
					branch_ = 2;
				}
			}
		}
		eepa_F_.rdof(0) = eepa_Fn;												//set the updated normal force component
		//
		// EEPA SHEAR FORCE ==============================================================================================================
		//
		DVect sforce(0.0);
		// dim holds the dimension (e.g., 2 for 2D and 3 for 3D)
		// Loop over the shear components (note: the 0 component is the normal component) and calculate the shear force.
		for (int i = 1; i < dim; ++i)
			sforce.rdof(i) = eepa_F_.dof(i) - trans.dof(i) * ks_;												// shear force update
		//
		// The canFail flag corresponds to whether or not the contact can undergo non-linear
		// force-displacement response. If the SOLVE ELASTIC command is given then the 
		// canFail state is set to FALSE. Otherwise it is always TRUE. 
		if (state->canFail_) {
			// Resolve sliding. This is the normal force multiplied by the coefficient of friction.
			double crit = fric_ * std::abs(eepa_Fn - f_min_);													//take absolute value - contact with very small overlap might result in negative value here due to rounding
			// The is the magnitude of the shear force.
			double sfmag = sforce.mag();
			// Sliding occurs when the magnitude of the shear force is greater than the critical value.
			if (sfmag > crit) {
				// Lower the shear force to the critical value for sliding.
				double rat = crit / sfmag;
				sforce *= rat;
				// Handle the slip_change event if one has been hooked up. Sliding has commenced.  
				if (!eepa_S_ && cmEvents_[fSlipChange] >= 0) {
					FArray<QVariant, 3> arg;
					QVariant p1;
					// Put a pointer to the contact in the array plus 0 to indicate slip has initiated.
					IContact * c = const_cast<IContact*>(state->getContact());
					TPtr<IThing> t(c->getIThing());
					p1.setValue(t);
					arg.push_back(p1);
					p1.setValue(0);
					arg.push_back(p1);
					IFishCallList *fi = const_cast<IFishCallList*>(state->getProgram()->findInterface<IFishCallList>());
					fi->setCMFishCallArguments(c, arg, cmEvents_[fSlipChange]);
				}
				eepa_S_ = true;
			}
			else {
				// Handle the slip_change event if one has been hooked up and
				// the contact was previously sliding. Sliding has ceased.  
				if (eepa_S_) {
					if (cmEvents_[fSlipChange] >= 0) {
						FArray<QVariant, 3> arg;
						QVariant p1;
						// Put a pointer to the contact in the array plus 1 to indicate slip has ceased.
						IContact * c = const_cast<IContact*>(state->getContact());
						TPtr<IThing> t(c->getIThing());
						p1.setValue(t);
						arg.push_back(p1);
						p1.setValue(1);
						arg.push_back(p1);
						IFishCallList *fi = const_cast<IFishCallList*>(state->getProgram()->findInterface<IFishCallList>());
						fi->setCMFishCallArguments(c, arg, cmEvents_[fSlipChange]);
					}
					eepa_S_ = false;
				}
			}
		}
		//
		// Set the shear components of the total force.
		for (int i = 1; i < dim; ++i)
			eepa_F_.rdof(i) = sforce.dof(i);
		//
		// EEPA ROLLING RESISTANCE ==============================================================================================================
		//
		DAVect res_M_old = res_M_;													// used in energy calculation only
		if (fr_ == 0.0) {											
			res_M_.fill(0.0);
		}
		else {
			DAVect angStiff(0.0);
			DAVect MomentInc(0.0);
#if DIM==3 
			angStiff.rx() = 0.0;
			angStiff.ry() = kr_;
#endif
			angStiff.rz() = kr_;
			MomentInc = ang * angStiff * (-1.0);
			res_M_ += MomentInc;
			if (state->canFail_) {
				// Account for bending strength
				double rmax = std::abs(fr_*(eepa_Fn - f_min_));					// Using the same normal force as in the shear limit
				double rmag = res_M_.mag();
				if (rmag > rmax) {
					double fac = rmax / rmag;
					res_M_ *= fac;
					res_S_ = true;
				}
				else {
					res_S_ = false;
				}
			}
		}
		//
		// EEPA STIFFNESS FOR TIME STEP UPDATES ============================================================================================
		//
		// set effective stiffness in the normal and shear directions for timestep calculation - kn_branch is based on current loading-unloading branch
		effectiveTranslationalStiffness_ = DVect2(kn_branch, ks_);
		//
		// EEPA DAMPING FORCE ==============================================================================================================
		//
		// Account for dashpot forces if the dashpot structure has been defined. 
		if (dpProps_) {
			dpProps_->dp_F_.fill(0.0);
			double vcn(0.0), vcs(0.0);
			// Calculate the damping coefficients. 
			setDampCoefficients(state->inertialMass_, &kn_branch, &ks_, &vcn, &vcs);				// this is based on the current normal stiffness kn_branch
			// First damp the shear components
			for (int i = 1; i < dim; ++i)
				dpProps_->dp_F_.rdof(i) = trans.dof(i) * (-1.0* vcs) / timestep;
			// Damp the normal component
			dpProps_->dp_F_.rx() -= trans.x() * vcn / timestep;
			//
			if (eepa_S_ && dpProps_->dp_mode_ == 1) {
				// Limit in shear if sliding.
				double dfn = dpProps_->dp_F_.rx();
				dpProps_->dp_F_.fill(0.0);
				dpProps_->dp_F_.rx() = dfn;
			}
			// Correct effective translational stiffness
			DVect2 correct(1.0);
			if (dpProps_->dp_nratio_)
				correct.rx() = sqrt(1.0 + dpProps_->dp_nratio_*dpProps_->dp_nratio_) - dpProps_->dp_nratio_;
			if (dpProps_->dp_sratio_)
				correct.ry() = sqrt(1.0 + dpProps_->dp_sratio_*dpProps_->dp_sratio_) - dpProps_->dp_sratio_;
			effectiveTranslationalStiffness_ /= (correct*correct);
		}
		
		//Compute energies if energy tracking has been enabled. 
		if (state->trackEnergy_) {
			assert(energies_);
			// calculate the strain energy increment in the normal direction				 
			energies_->estrain_ -= 0.5*(eepa_F_old.dof(0) + eepa_Fn)*trans.dof(0);// under loading, trans.x < 0, hence the negative sign
			if (ks_) {
				DVect u_s_elastic = trans;										// set the elastic displacement increment equal to the total displacement increment
				u_s_elastic.rx() = 0.0;											// set the normal component to zero: u_s_elatic = [0, trans_shear_1, trans_shear_2]
				DVect shearF = eepa_F_;											// set the shear force equal to the total EEPA force (including the normal component)
				shearF.rx() = 0.0;												// set the normal component to zero: shearF = [0, shear_force_1, shear_force_2]
				eepa_F_old.rx() = 0.0;											// set normal component of the previous force equal to zero
				DVect avg_F_s = (shearF + eepa_F_old)*0.5;						// average shear force vector
				if (eepa_S_) {													// if sliding, calculate the slip energy and accumulate it
					DVect u_s_total = u_s_elastic;								// total shear displacement increment
					u_s_elastic = (shearF - eepa_F_old) / ks_;					// elastic shear displacement increment
					energies_->eslip_ -= std::min(0.0, (avg_F_s | (u_s_total + u_s_elastic))); // where (u_s_total + u_s_elatic) is the slip displacment increment (due to the sign convention, the terms are added up)
				}
				energies_->estrain_ -= avg_F_s | u_s_elastic;
			}
			// Add the rolling resistance energy contributions.
			energies_->errstrain_ = 0.0;
			if (kr_) {
				energies_->errstrain_ = 0.5*res_M_.mag2() / kr_;
				if (res_S_) {
					// If sliding calculate the slip energy and accumulate it.
					DAVect avg_M = (res_M_ + res_M_old)*0.5;
					DAVect t_s_el = (res_M_ - res_M_old) / kr_;
					energies_->errslip_ -= std::min(0.0, (avg_M | (ang + t_s_el)));
				}
			}
			// Calculate damping energy (accumulated) if the dashpots are active. 
			if (dpProps_) {
				energies_->edashpot_ -= dpProps_->dp_F_ | trans;
			}
		}

		// This is just a sanity check to ensure, in debug mode, that the force isn't wonky. 
		assert(eepa_F_ == eepa_F_);
		return true;
	}

	void ContactModelEEPA_LIN::setForce(const DVect &v, IContact *c) {
		//
	}

	DVect ContactModelEEPA_LIN::getForce(const IContactMechanical *) const {
		DVect ret(eepa_F_);																				
		if (dpProps_)
			ret += dpProps_->dp_F_;																		
		return ret;
	}

	DAVect ContactModelEEPA_LIN::getMomentOn1(const IContactMechanical *c) const {
		DVect force = getForce(c);
		DAVect ret(res_M_);
		c->updateResultingTorqueOn1Local(force, &ret);
		return ret;
	}

	DAVect ContactModelEEPA_LIN::getMomentOn2(const IContactMechanical *c) const {
		DVect force = getForce(c);
		DAVect ret(res_M_);
		c->updateResultingTorqueOn2Local(force, &ret);
		return ret;
	}

	void ContactModelEEPA_LIN::setDampCoefficients(const double &mass, double *kn, double *ks_, double *vcn, double *vcs) {
		*vcn = dpProps_->dp_nratio_ * 2.0 * sqrt(mass*(*kn));
		*vcs = dpProps_->dp_sratio_ * 2.0 * sqrt(mass*(*ks_));
	}

} // namespace cmodelsxd
// EoF