// eepa_nl.h

#include "contactmodel/src/contactmodelmechanical.h"

#ifdef EEPA_NL_LIB
#  define EEPA_NL_EXPORT EXPORT_TAG
#elif defined(NO_MODEL_IMPORT)
#  define EEPA_NL_EXPORT
#else
#  define EEPA_NL_EXPORT IMPORT_TAG
#endif

namespace cmodelsxd {
	using namespace itasca;

	class ContactModelEEPA_NL : public ContactModelMechanical {
	public:
		// Constructor: Set default values for contact model properties.
		EEPA_NL_EXPORT ContactModelEEPA_NL();
		// Destructor, called when contact is deleted: free allocated memory, etc.
		EEPA_NL_EXPORT virtual ~ContactModelEEPA_NL();
		// Contact model name (used as keyword for commands and FISH).
		virtual QString  getName() const { return "eepa_nl"; }
		// The index provides a quick way to determine the type of contact model.
		// Each type of contact model in PFC must have a unique index; this is assigned
		// by PFC when the contact model is loaded. This index should be set to -1
		virtual void     setIndex(int i) { index_ = i; }
		virtual int      getIndex() const { return index_; }
		// Contact model version number (e.g., MyModel05_1). The version number can be
		// accessed during the save-restore operation (within the archive method,
		// testing {stream.getRestoreVersion() == getMinorVersion()} to allow for 
		// future modifications to the contact model data structure.
		virtual uint     getMinorVersion() const;
		// Copy the state information to a newly created contact model.
		// Provide access to state information, for use by copy method.
		virtual void     copy(const ContactModel *c);
		// Provide save-restore capability for the state information.
		virtual void     archive(ArchiveStream &);
		// Enumerator for the properties.
		enum PropertyKeys {
			kwShear = 1
			, kwPoiss
			, kwFric
			, kwEepaF
			, kwEepaS
			, kwRGap
			, kwDpNRatio
			, kwDpSRatio
			, kwDpMode
			, kwDpF
			, kwResFric
			, kwResMoment
			, kwResS
			, kwOverlapMax
			, kwPlasRat
			, kwLuExp
			, kwPullOff
			, kwAdhExp
			, kwSurfAdh
			, kwKsFac		
			, kwFmin
		};
		// Contact model property names in a comma separated list. The order corresponds with
		// the order of the PropertyKeys enumerator above. One can visualize any of these 
		// properties in PFC automatically. 
		virtual QString  getProperties() const {
			return "eepa_shear"
				",eepa_poiss"
				",fric"
				",eepa_force"
				",eepa_slip"
				",rgap"
				",dp_nratio"
				",dp_sratio"
				",dp_mode"
				",dp_force"
				",rr_fric"
				",rr_moment"
				",rr_slip"
				",overlap_max"
				",plas_ratio"
				",lu_exp"
				",pull_off"
				",adh_exp"
				",surf_adh"
				",ks_fac"
				",f_min";			
		}
		// Enumerator for the energies.
		enum EnergyKeys {
			kwEStrain = 1
			, kwERRStrain
			, kwESlip
			, kwERRSlip
			, kwEDashpot
		};
		// Contact model energy names in a comma separated list. The order corresponds with
		// the order of the EnergyKeys enumerator above. 
		virtual QString  getEnergies() const {
			return "energy-strain-eepa"
				",energy-rrstrain"
				",energy-slip"
				",energy-rrslip"
				",energy-dashpot";
		}
		// Returns the value of the energy (base 1 - getEnergy(1) returns the estrain energy).
		virtual double   getEnergy(uint i) const;
		// Returns whether or not each energy is accumulated (base 1 - getEnergyAccumulate(1) 
		// returns whether or not the estrain energy is accumulated which is false).
		virtual bool     getEnergyAccumulate(uint i) const;
		// Set an energy value (base 1 - setEnergy(1) sets the estrain energy).
		virtual void     setEnergy(uint i, const double &d); // Base 1
		// Activate the energy. This is only called if the energy tracking is enabled. 
		virtual void     activateEnergy() { if (energies_) return; energies_ = NEWC(Energies()); }
		// Returns whether or not the energy tracking has been enabled for this contact.
		virtual bool     getEnergyActivated() const { return (energies_ != 0); }

		// Enumerator for contact model related FISH callback events. 
		enum FishCallEvents {
			fActivated = 0
			, fSlipChange
		};
		// Contact model FISH callback event names in a comma separated list. The order corresponds with
		// the order of the FishCallEvents enumerator above. 
		virtual QString  getFishCallEvents() const {
			return
				"contact_activated"
				",slip_change";
		}

		// Return the specified contact model property.
		virtual QVariant getProperty(uint i, const IContact *) const;
		// The return value denotes whether or not the property corresponds to the global
		// or local coordinate system (TRUE: global system, FALSE: local system). The
		// local system is the contact-plane system (nst) defined as follows.
		// If a vector V is expressed in the local system as (Vn, Vs, Vt), then V is
		// expressed in the global system as {Vn*nc + Vs*sc + Vt*tc} where where nc, sc
		// and tc are unit vectors in directions of the nst axes.
		// This is used when rendering contact model properties that are vectors.
		virtual bool     getPropertyGlobal(uint i) const;
		// Set the specified contact model property, ensuring that it is of the correct type
		// and within the correct range --- if not, then throw an exception.
		// The return value denotes whether or not the update has affected the timestep
		// computation (by having modified the translational or rotational tangent stiffnesses).
		// If true is returned, then the timestep will be recomputed.
		virtual bool     setProperty(uint i, const QVariant &v, IContact *);
		// The return value denotes whether or not the property is read-only
		// (TRUE: read-only, FALSE: read-write).
		virtual bool     getPropertyReadOnly(uint i) const;

		// The return value denotes whether or not the property is inheritable
		// (TRUE: inheritable, FALSE: not inheritable). Inheritance is provided by
		// the endPropertyUpdated method.
		virtual bool     supportsInheritance(uint i) const;
		// Return whether or not inheritance is enabled for the specified property.
		virtual bool     getInheritance(uint i) const { assert(i < 32); quint32 mask = to<quint32>(1 << i);  return (inheritanceField_ & mask) ? true : false; }
		// Set the inheritance flag for the specified property.
		virtual void     setInheritance(uint i, bool b) { assert(i < 32); quint32 mask = to<quint32>(1 << i);  if (b) inheritanceField_ |= mask;  else inheritanceField_ &= ~mask; }

		// Prepare for entry into ForceDispLaw. The validate function is called when:
		// (1) the contact is created, (2) a property of the contact that returns a true via
		// the setProperty method has been modified and (3) when a set of cycles is executed
		// via the {cycle N} command.
		// Return value indicates contact activity (TRUE: active, FALSE: inactive).
		virtual bool    validate(ContactModelMechanicalState *state, const double &timestep);
		// The endPropertyUpdated method is called whenever a surface property (with a name
		// that matches an inheritable contact model property name) of one of the contacting
		// pieces is modified. This allows the contact model to update its associated
		// properties. The return value denotes whether or not the update has affected
		// the time step computation (by having modified the translational or rotational
		// tangent stiffnesses). If true is returned, then the time step will be recomputed.  
		virtual bool    endPropertyUpdated(const QString &name, const IContactMechanical *c);
		// The forceDisplacementLaw function is called during each cycle. Given the relative
		// motion of the two contacting pieces (via
		//   state->relativeTranslationalIncrement_ (Ddn, Ddss, Ddst)
		//   state->relativeAngularIncrement_       (Dtt, Dtbs, Dtbt)
		//     Ddn  : relative normal-displacement increment, Ddn > 0 is opening
		//     Ddss : relative  shear-displacement increment (s-axis component)
		//     Ddst : relative  shear-displacement increment (t-axis component)
		//     Dtt  : relative twist-rotation increment
		//     Dtbs : relative  bend-rotation increment (s-axis component)
		//     Dtbt : relative  bend-rotation increment (t-axis component)
		//       The relative displacement and rotation increments:
		//         Dd = Ddn*nc + Ddss*sc + Ddst*tc
		//         Dt = Dtt*nc + Dtbs*sc + Dtbt*tc
		//       where nc, sc and tc are unit vectors in direc. of the nst axes, respectively.
		//       [see {Table 1: Contact State Variables} in PFC Model Components:
		//       Contacts and Contact Models: Contact Resolution]
		// ) and the contact properties, this function must update the contact force and
		// moment.
		//   The force_ is acting on piece 2, and is expressed in the local coordinate system
		//   (defined in getPropertyGlobal) such that the first component positive denotes
		//   compression. If we define the moment acting on piece 2 by Mc, and Mc is expressed
		//   in the local coordinate system (defined in getPropertyGlobal), then we must use the getMechanicalContact()->updateResultingTorquesLocal(...) method to 
		// get the total moment. 
		// The return value indicates the contact activity status (TRUE: active, FALSE:
		// inactive) during the next cycle.
		// Additional information:
		//   * If state->activated() is true, then the contact has just become active (it was
		//     inactive during the previous time step).
		//   * Fully elastic behavior is enforced during the SOLVE ELASTIC command by having
		//     the forceDispLaw handle the case of {state->canFail_ == true}.
		virtual bool    forceDisplacementLaw(ContactModelMechanicalState *state, const double &timestep);
		// The getEffectiveXStiffness functions return the translational and rotational
		// tangent stiffnesses used to compute a stable time step. When a contact is sliding,
		// the translational tangent shear stiffness is zero (but this stiffness reduction
		// is typically ignored when computing a stable time step). If the contact model
		// includes a dashpot, then the translational stiffnesses must be increased (see
		// Potyondy (2009)).
		//   [Potyondy, D. 'Stiffness Matrix at a Contact Between Two Clumps,' Itasca
		//   Consulting Group, Inc., Minneapolis, MN, Technical Memorandum ICG6863-L,
		//   December 7, 2009.]
		virtual DVect2  getEffectiveTranslationalStiffness() const { return effectiveTranslationalStiffness_; }
		virtual DAVect  getEffectiveRotationalStiffness() const { return effectiveRotationalStiffness_; }

		// Return a new instance of the contact model. This is used in the CMAT
		// when a new contact is created. 
		virtual ContactModelEEPA_NL *clone() const { return NEWC(ContactModelEEPA_NL()); }
		// The getActivityDistance function is called by the contact-resolution logic when
		// the CMAT is modified. Return value is the activity distance used by the
		// checkActivity function.
		virtual double              getActivityDistance() const { return rgap_; }							
		// The isOKToDelete function is called by the contact-resolution logic when...
		// Return value indicates whether or not the contact may be deleted.
		// If TRUE, then the contact may be deleted when it is inactive.
		// If FALSE, then the contact may not be deleted (under any condition).
		virtual bool                isOKToDelete() const { return !isBonded(); }
		// Zero the forces and moments stored in the contact model. This function is called
		// when the contact becomes inactive.
		virtual void                resetForcesAndMoments() {
			eepa_F(DVect(0.0)); dp_F(DVect(0.0));
			res_M(DAVect(0.0));																			
			if (energies_) {
				//energies_->estrain_ = 0.0;
				energies_->errstrain_ = 0.0;
			}
			Moverlap_ = 0.0;																// reset the history dependend variables
			Poverlap_exp_ = 0.0;
			branch_ = 1;
		}
		virtual void     setForce(const DVect &v, IContact *c);
		virtual void     setArea(const double &d) { }
		// The checkActivity function is called by the contact-resolution logic when...
		// Return value indicates contact activity (TRUE: active, FALSE: inactive).
		// A contact with the arrlinear model is active if the surface gap is less than
		// or equal to the attraction range (a_d0_).
		virtual bool     checkActivity(const double &gap) { return  gap <= rgap_; }							
		
		// Returns the sliding state (FALSE is returned if not implemented).
		virtual bool     isSliding() const { return eepa_S_; }
		// Returns the bonding state (FALSE is returned if not implemented).
		virtual bool     isBonded() const { return false; }

		// Both of these methods are called only for contacts with facets where the wall 
		// resolution scheme is set the full. In such cases one might wish to propagate 
		// contact state information (e.g., shear force) from one active contact to another. 
		// See the Faceted Wall section in the documentation. 

		// Return the total force that the contact model holds.
		virtual DVect    getForce(const IContactMechanical *) const;

		// Return the total moment on 1 that the contact model holds
		virtual DAVect   getMomentOn1(const IContactMechanical *) const;

		// Return the total moment on 1 that the contact model holds
		virtual DAVect   getMomentOn2(const IContactMechanical *) const;

		// Methods to get and set properties. 
		const double & shear() const { return shear_; }
		void           shear(const double &d) { shear_ = d; }
		const double & poiss() const { return poiss_; }
		void           poiss(const double &d) { poiss_ = d; }
		const double & fric() const { return fric_; }
		void           fric(const double &d) { fric_ = d; }
		const DVect &  eepa_F() const { return eepa_F_; }
		void           eepa_F(const DVect &f) { eepa_F_ = f; }
		bool           eepa_S() const { return eepa_S_; }
		void           eepa_S(bool b) { eepa_S_ = b; }
		const double & rgap() const { return rgap_; }
		void           rgap(const double &d) { rgap_ = d; }
		const double & Moverlap() const { return Moverlap_; }
		void           Moverlap(const double &d) { Moverlap_ = d; }
		const double & Poverlap_exp() const { return Poverlap_exp_; }
		void           Poverlap_exp(const double &d) { Poverlap_exp_ = d; }
		const double & plas_ratio() const { return plas_ratio_; }
		void           plas_ratio(const double &d) { plas_ratio_ = d; }
		const double & lu_exp() const { return lu_exp_; }
		void		   lu_exp(const double &d) { lu_exp_ = d; }
		const double & pull_off() const { return pull_off_; }
		void           pull_off(const double &d) { pull_off_ = d; }
		const double & adh_exp() const { return adh_exp_; }
		void	       adh_exp(const double &d) { adh_exp_ = d; }
		const double & surf_adh() const { return surf_adh_; }
		void           surf_adh(const double &d) { surf_adh_ = d; }
		const double & lu_exp_inv() const { return lu_exp_inv_; }
		void           lu_exp_inv(const double &d) { lu_exp_inv_ = d; }
		const double & k1() const { return k1_; }
		void           k1(const double &d) { k1_ = d; }
		const double & k2() const { return k2_; }
		void           k2(const double &d) { k2_ = d; }
		const double & kadh() const { return kadh_; }
		void           kadh(const double &d) { kadh_ = d; }
		const double & ks() const { return ks_; }
		void           ks(const double &d) { ks_ = d; }
		const double & r_hertz() const { return r_hertz_; }
		void           r_hertz(const double &d) { r_hertz_ = d; }
		const double & ks_fac() const { return ks_fac_; }
		void           ks_fac(const double &d) { ks_fac_ = d; }
		const double & f_min() const { return f_min_; }
		void           f_min(const double &d) { f_min_ = d; }
		const int    & branch() const { return branch_; }
		void           branch(const int &d) { branch_ = d; }
		const double & rbar_square() const { return rbar_square_; }
		void           rbar_square(const double &d) { rbar_square_ = d; }

		bool     hasDamping() const { return dpProps_ ? true : false; }
		double   dp_nratio() const { return (hasDamping() ? (dpProps_->dp_nratio_) : 0.0); }
		void     dp_nratio(const double &d) { if (!hasDamping()) return; dpProps_->dp_nratio_ = d; }
		double   dp_sratio() const { return hasDamping() ? dpProps_->dp_sratio_ : 0.0; }
		void     dp_sratio(const double &d) { if (!hasDamping()) return; dpProps_->dp_sratio_ = d; }
		int      dp_mode() const { return hasDamping() ? dpProps_->dp_mode_ : -1; }
		void     dp_mode(int i) { if (!hasDamping()) return; dpProps_->dp_mode_ = i; }
		DVect    dp_F() const { return hasDamping() ? dpProps_->dp_F_ : DVect(0.0); }
		void     dp_F(const DVect &f) { if (!hasDamping()) return; dpProps_->dp_F_ = f; }

		bool    hasEnergies() const { return energies_ ? true : false; }
		double  estrain() const { return hasEnergies() ? energies_->estrain_ : 0.0; }
		void    estrain(const double &d) { if (!hasEnergies()) return; energies_->estrain_ = d; }
		double  errstrain() const { return hasEnergies() ? energies_->errstrain_ : 0.0; }
		void    errstrain(const double &d) { if (!hasEnergies()) return; energies_->errstrain_ = d; }
		double  eslip() const { return hasEnergies() ? energies_->eslip_ : 0.0; }
		void    eslip(const double &d) { if (!hasEnergies()) return; energies_->eslip_ = d; }
		double  errslip() const { return hasEnergies() ? energies_->errslip_ : 0.0; }
		void    errslip(const double &d) { if (!hasEnergies()) return; energies_->errslip_ = d; }
		double  edashpot() const { return hasEnergies() ? energies_->edashpot_ : 0.0; }
		void    edashpot(const double &d) { if (!hasEnergies()) return; energies_->edashpot_ = d; }
		
		uint inheritanceField() const { return inheritanceField_; }
		void inheritanceField(uint i) { inheritanceField_ = i; }

		const DVect2 & effectiveTranslationalStiffness()  const { return effectiveTranslationalStiffness_; }
		void           effectiveTranslationalStiffness(const DVect2 &v) { effectiveTranslationalStiffness_ = v; }
		const DAVect & effectiveRotationalStiffness()  const { return effectiveRotationalStiffness_; }
		void           effectiveRotationalStiffness(const DAVect &v) { effectiveRotationalStiffness_ = v; }

		// Rolling resistance methods
		const double & res_fric() const { return res_fric_; }
		void           res_fric(const double &d) { res_fric_ = d; }
		const DAVect & res_M() const { return res_M_; }
		void           res_M(const DAVect &f) { res_M_ = f; }
		bool           res_S() const { return res_S_; }
		void           res_S(bool b) { res_S_ = b; }
		const double & kr() const { return kr_; }
		void           kr(const double &d) { kr_ = d; }
		const double & fr() const { return fr_; }
		void           fr(const double &d) { fr_ = d; }

	private:
		// Index - used internally by PFC. Should be set to -1 in the cpp file. 
		static int index_;

		// Structure to store the energies. 
		struct Energies {
			Energies() : estrain_(0.0), errstrain_(0.0), eslip_(0.0), errslip_(0.0), edashpot_(0.0) {}
			double estrain_;   // elastic energy stored in EEPA group 
			double errstrain_; // elastic energy stored in rolling resistance group
			double eslip_;     // work dissipated by friction 
			double errslip_;   // work dissipated by rolling resistance friction 
			double edashpot_;  // work dissipated by dashpots
		};

		// Structure to store dashpot quantities. 
		struct dpProps {
			dpProps() : dp_nratio_(0.0), dp_sratio_(0.0), dp_mode_(0), dp_F_(DVect(0.0)) {}
			double dp_nratio_;		// normal viscous critical damping ratio
			double dp_sratio_;		// shear  viscous critical damping ratio
			int    dp_mode_;		// for viscous mode (0-1) 0 = no cut-offs;    1 = shear damp cut-off when sliding
			DVect  dp_F_;			// Force in the dashpots
		};

		bool   updateEndStiffCoef(const IContactMechanical *con);
		bool   updateEndFric(const IContactMechanical *con);
		bool   updateEndResFric(const IContactMechanical *con);
		void   updateStiffCoef(const IContactMechanical *con);
		void   updateEffectiveStiffness(ContactModelMechanicalState *state);
		void   setDampCoefficients(const double &mass, double *kn_tangent, double *ks_tangent, double *vcn, double *vcs);

		// Contact model inheritance fields.
		quint32 inheritanceField_;

		// Effective translational stiffness.
		DVect2  effectiveTranslationalStiffness_;
		DAVect  effectiveRotationalStiffness_;      // (Twisting,Bending,Bending) Rotational stiffness (twisting always 0)

		// EEPA model properties 
		double      shear_;				// Shear modulus
		double      poiss_;				// Poisson's ratio
		double      fric_;				// Coulomb friction coefficient
		DVect       eepa_F_;			// Force carried in the eepa model
		bool        eepa_S_;			// The current slip state
		double      rgap_;				// Reference gap 
		dpProps *   dpProps_;			// The viscous properties
		double		Moverlap_;			// Maximum overlap - updated as it evolves
		double		Poverlap_exp_;	    // Plastic overlap raised to the power lu_exp - updated as it evolves
		double		plas_ratio_;		// plasticity ratio
		double		lu_exp_;			// load-unload exponent anlong the k1 and k2 brances
		double		pull_off_;			// constant pull-off force
		double		adh_exp_;			// adhesive branch exponent
		double		surf_adh_;			// surface adhesion energy
		double		lu_exp_inv_;		// inverse of load-unloading exponent
		double      k1_;				// k1 branch stiffness
		double		k2_;				// k2 branch stiffness
		double      kadh_;				// adhesive branch stiffness - updated as the maximum overlap (and plastic overlap) increases
		double      ks_;				// shear stiffness
		double      r_hertz_;    		// effective contact radius = (R1*R2)/(R1 + R2)
		double      ks_fac_;			// shear stiffness scaling factor
		double		f_min_;				// minimum adhesion force limit
		int         branch_;			// 1,2 or 3 for the current branch
		
		// rolling resistance properties
		double res_fric_;       // rolling friction coefficient
		DAVect res_M_;          // moment (bending only)         
		bool   res_S_;          // The current rolling resistance slip state
		double kr_;             // bending rotational stiffness (read-only, calculated internaly) 
		double fr_;             // rolling friction coefficient (rbar*res_fric_) (calculated internaly, not a property)
		double rbar_square_;	// curvature expression used in calculating the rolling stiffness - stays constant for  given contact where kr_ = rbar_square*ks_tangent

		Energies *   energies_; // The energies

	};
} // namespace cmodelsxd
// EoF